from time import perf_counter_ns
from typing import Self

class Timer:
    __slots__ = ("name", "_execution_time", "_start_time", "count")
    name: str
    count: int
    _execution_time: int


    def __init__(self, name: str) -> None:
        self.name = name
        self._execution_time = 0
        self.count = 0

    def get_ms(self) -> float:
        return self.get_ns() / 1_000_000

    def get_ns(self) -> int:
        assert not hasattr(self, "_start_time"), "Timer not stopped yet"
        return self._execution_time

    def __enter__(self) -> Self:
        assert not hasattr(self, "_start_time"), "Timer not stopped yet"
        self._start_time = perf_counter_ns()
        return self

    def __exit__(self, exc_type: object, exc_val: object, exc_tb: object) -> None:
        self._execution_time += perf_counter_ns() - self._start_time
        del self._start_time
        self.count += 1

    def __str__(self) -> str:
        return f"{self.name}: {self.get_ms() / self.count:.4} ms"



import sys
from collections.abc import Buffer

import qoi  # type: ignore[import-untyped]
import qoi_rs
import numpy
from PIL import Image
from .timer import Timer


def qoi_encode(image: Image.Image) -> bytes:
    return qoi.encode(numpy.asarray(image))


def qoi_rs_encode(image: Image.Image) -> bytes:
    return qoi_rs.encode_pillow(image)


def main(args: list[str]) -> int | str:
    timers = [(Timer(fun.__name__), fun) for fun in (qoi_encode, qoi_rs_encode)]

    for _ in range(10):
        timers.reverse()
        for image_path in args:
            qoi_images = set()
            image = Image.open(image_path)
            image.load()
            for (timer, fun) in timers:
                with timer:
                    qoi_img = fun(image)
                assert isinstance(qoi_img, Buffer)
                if not _:
                    decoded = qoi_rs.decode(qoi_img)
                    qoi_images.add((decoded.width, decoded.height, decoded.mode, decoded.colour_space, decoded.data))
            image.close()
            del image
            if not _:
                assert len(qoi_images) == 1, f"{qoi_images}"

    print(f"Benchmarked with {len(args)} images")
    for (timer, _) in timers:
        print(timer)

    for (timer, _) in timers:
        assert timer.count == len(args) * 10, f"{timer.name} has wrong count {timer.count}"

    return 0


sys.exit(main(sys.argv[1:]))



from collections.abc import Sequence, Buffer
from typing import Literal, Protocol, runtime_checkable, Never, TYPE_CHECKING

__all__ = "Data", "Image"

if TYPE_CHECKING:
    try:
        from PIL.Image import Image as PillowImage
    except ModuleNotFoundError:
        PillowImage = Never
else:
    PillowImage = Never


class ArrowArrayExportable(Protocol):
    # SEE: https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html#protocol-typehints
    def __arrow_c_array__(
        self,
        requested_schema: object | None = None
    ) -> tuple[object, object]:
        pass


type Data = (
    Sequence[tuple[int, int, int]]
    | Sequence[tuple[int, int, int, int]]
    | Sequence[int]
    | bytes
    | bytearray
    | Buffer
    | ArrowArrayExportable
    | PillowImage
)

type ColourSpace = Literal["SRGB", "linear"]

type Mode = Literal["RGB", "RGBA"]

@runtime_checkable
class Image(Protocol):
    @property
    def width(self) -> int: pass
    @property
    def height(self) -> int: pass
    @property
    def data(self) -> bytes: pass
    @property
    def channels(self) -> int: pass
    @property
    def colour_space(self) -> ColourSpace: pass
    @property
    def mode(self) -> Mode: pass



from collections.abc import Buffer
from typing import TYPE_CHECKING

from . import types

__all__ = "encode", "decode", "encode_pillow", "decode_pillow"

if TYPE_CHECKING:

    def encode(
        data: types.Data,
        /, *,
        width: int,
        height: int,
        colour_space: types.ColourSpace = None,
        mode: types.Mode = None,
    ) -> bytes:
        pass

    def decode(data: Buffer, /) -> types.Image:
        pass

else:
    from ._qoi import encode, decode


def encode_pillow(
    image: types.PillowImage,
    /, *,
    colour_space: types.ColourSpace = None,
    mode: types.Mode = None,
):
    return encode(
        image.tobytes(), # TODO: use image directly: https://github.com/python-pillow/Pillow/issues/8329
        width=image.width,
        height=image.height,
        colour_space=colour_space,
        mode=mode or image.mode,
    )


def decode_pillow(data: Buffer) -> types.PillowImage:
    from PIL import Image
    image = decode(data)
    return Image.frombytes(
        image.mode,
        (image.width, image.height),
        image.data,
    )



#!/usr/bin/env -S uv run --reinstall-package qoi-rs --script
# /// script
# dependencies = [
#   "qoi-rs",
#   "pillow",
# ]
# [tool.uv.sources]
# qoi-rs = { path = "." }
# ///

from collections.abc import Buffer
from pathlib import Path

import qoi_rs

from PIL import Image

IMAGES_DIR = Path(__file__).absolute().parent / "qoi_test_images"


for file in "dice", "testcard":
    qoi_image_path = IMAGES_DIR / f"{file}.qoi"
    png_image_path = IMAGES_DIR / f"{file}.png"

    image = qoi_rs.decode(qoi_image_path.read_bytes())
    assert isinstance(image, qoi_rs.types.Image)
    png_image = Image.open(png_image_path)
    pil_qoi_image = Image.open(qoi_image_path)

    assert image.width == png_image.width == pil_qoi_image.width
    assert image.height == png_image.height == pil_qoi_image.height
    assert image.mode == pil_qoi_image.mode
    assert isinstance(image.data, Buffer)
    assert image.mode in {"RGB", "RGBA"}

    for img in png_image, pil_qoi_image:
        encoded = qoi_rs.encode(img.getdata(), width=img.width, height=img.height)
        assert encoded == qoi_rs.encode(img, width=img.width, height=img.height, mode=img.mode)  # type: ignore[arg-type]
        assert encoded == qoi_rs.encode(img.tobytes(), width=img.width, height=img.height)
        assert encoded == qoi_rs.encode_pillow(img)
        assert isinstance(encoded, Buffer)
        assert type(encoded) is bytes
        decoded = qoi_rs.decode(encoded)
        assert image.width == decoded.width
        assert image.height == decoded.height
        assert image.colour_space == decoded.colour_space
        assert image.mode == decoded.mode, f"{file}: {image.mode} != {decoded.mode}"
        assert image.channels == image.channels
        assert type(image) is type(decoded)
        assert len(image.data) == len(decoded.data), f"{file}: {len(image.data)} != {len(decoded.data)}"
        assert image.data == decoded.data, f"{file}"
        assert image == decoded, f"{file}"

        assert tuple(png_image.getdata()) == tuple(qoi_rs.decode_pillow(encoded).getdata())

    encoded = qoi_rs.encode(image.data, width=image.width, height=image.height, colour_space="linear")
    assert qoi_rs.decode(encoded).colour_space == "linear"
    encoded = qoi_rs.encode(image.data, width=image.width, height=image.height, colour_space="SRGB")
    assert qoi_rs.decode(encoded).colour_space == "SRGB"

    for img in png_image, pil_qoi_image:
        encoded = qoi_rs.encode_pillow(img, colour_space="linear")
        assert qoi_rs.decode(encoded).colour_space == "linear"
        encoded = qoi_rs.encode_pillow(img, colour_space="SRGB")
        assert qoi_rs.decode(encoded).colour_space == "SRGB"

    try:
        qoi_rs.encode(image.data, width=image.width, height=image.height, colour_space="test")  # type: ignore[arg-type]
    except ValueError as err:
        assert len(err.args) == 1
        assert "invalid colour space" in err.args[0]
    else:
        assert False

    encoded = qoi_rs.encode(image.data, width=image.width, height=image.height, mode="RGB")
    assert qoi_rs.decode(encoded).mode == "RGB"
    encoded = qoi_rs.encode(image.data, width=image.width, height=image.height, mode="RGBA")
    assert qoi_rs.decode(encoded).mode == "RGBA"

    for img in png_image, pil_qoi_image:
        encoded = qoi_rs.encode_pillow(img, mode="RGB")
        assert qoi_rs.decode(encoded).mode == "RGB"
        encoded = qoi_rs.encode_pillow(img, mode="RGBA")
        assert qoi_rs.decode(encoded).mode == "RGBA"

    try:
        qoi_rs.encode(image.data, width=image.width, height=image.height, mode="test")  # type: ignore[arg-type]
    except ValueError as err:
        assert len(err.args) == 1
        assert "invalid mode" in err.args[0]
    else:
        assert False

    try:
        qoi_rs.encode(image.data, width=image.width + 1, height=image.height)
    except ValueError as err:
        assert len(err.args) == 1
        assert err.args[0].startswith("invalid image length: ")
    else:
        assert False

    try:
        qoi_rs.encode(image.data, width=image.width, height=image.height + 1)
    except ValueError as err:
        assert len(err.args) == 1
        assert err.args[0].startswith("invalid image length: ")
    else:
        assert False

    png_image.close()
    pil_qoi_image.close()

try:
    qoi_rs.encode(b"", width=1, height=1)
except ValueError as err:
    assert len(err.args) == 1
    assert "invalid number of channels: 0" == err.args[0]
else:
    assert False

try:
    qoi_rs.decode(b"")
except ValueError as err:
    assert len(err.args) == 1
    assert "unexpected input buffer end while decoding" == err.args[0]
else:
    assert False


assert qoi_rs.decode(qoi_rs.encode([(1, 2, 3)] * 12, width=4, height=3)).mode == "RGB"
assert qoi_rs.decode(qoi_rs.encode([(1, 2, 3, 4)] * 12, width=4, height=3)).mode == "RGBA"
try:
    qoi_rs.encode([(1, 2, 3, 4)] * 9, width=4, height=3)
except ValueError as err:
    assert len(err.args) == 1
    assert "got 9 pixels, image can't be 4x3" == err.args[0]
else:
    assert False


assert qoi_rs.decode(qoi_rs.encode([(1, 2, 3, 4)], width=1, height=1, colour_space="linear")).colour_space == "linear"
assert qoi_rs.decode(qoi_rs.encode([(1, 2, 3, 4)], width=1, height=1, colour_space="SRGB")).colour_space == "SRGB"
try:
    qoi_rs.encode([(1, 2, 3, 4)], width=1, height=1, colour_space="ÄÖÜ")  # type: ignore[arg-type]
except ValueError as err:
    assert len(err.args) == 1
    assert 'invalid colour space, needs to be one of ["linear", "SRGB"]' == err.args[0]
else:
    assert False
